#define detect                       3     // вход детектор нуля (спад за 200 мкс до нуля, фронт 200 мкс после нуля)
#define triac                        5     // выход на симистор
#define BUTTON_SET                   7     //Вывод кнопки "SET" (подтянут вверх, кнопка одной ногой на GND)
#define BUTTON_MINUS                 8     //Вывод кнопки "-" (подтянут вверх, кнопка одной ногой на GND)
#define BUTTON_PLUS                  9     //Вывод кнопки "+" (подтянут вверх, кнопка одной ногой на GND)
#define ACCELERATOR                  10    //Вывод разгона (подтянут вверх, активен когда замкнут на GND, при 
                                           // отпускании вызывает напряжение из памяти)
#define STOP                         11    //Вывод стоп (подтянут вверх, активен когда замкнут на GND, имеет приоритет
                                           // над разгоном, при отпускании и активном разгоне - переход в разгон,
                                           // при отпускании и неактивном разгоне - переход в 0 Вольт)
#define EE_INIT_ADDR                 0
#define EE_UST_U_ADDR                1
#define DEFAULT_UST_U                100   // значение напряжения дефолтное


#include <Wire.h>
#include <LiquidCrystal_I2C.h> //https://github.com/marcoschwartz/LiquidCrystal_I2C
#include <EEPROM.h>

LiquidCrystal_I2C lcd(0x3F, 16, 2);

volatile unsigned long Uism = 0;          //мгновенные значения напряжения
volatile int cntr = 0;                    // счетчик в обработчике прерывания
volatile unsigned long Usumm = 0;         // переменная для хранения суммы квадратов

float real_U = 0;                         //переменная расчета RMS напряжения
float sqrtUsum = 0;                       //переменная расчете квадратного корня
float ust_U = 0;                          //установленное напряжение
float step_ = 0;                          //шаг регулировки угла
unsigned long previousMillis = 0;
unsigned long displayMillis = 0;
unsigned long reg_time = 0;
unsigned long countMillis = 0;
byte set = 0;                             //переменная включения разгона

int angle_U = 600;                        // Угол ставим в положение при котором устойчивый ноль на выходе (нужно подбирать)
int t = 0;
int zero = 0;

float buffer_ust_U = 0;                   //напряжение уставка
int BUTTON_PLUS_press_time = 0;
unsigned long time1_press = 0;
int BUTTON_MINUS_press_time = 0;
unsigned long time2_press = 0;
int BUTTON_SET_press_time = 0;
unsigned long time3_press = 0;
int ACCELERATOR_press_time = 0;
unsigned long time4_press = 0;
int STOP_press_time = 0;
unsigned long time5_press = 0;
int BUTTON_STOP_press_time = 0;
unsigned long time6_press = 0;

// Фильтр НЧ 1 порядка (100 / 1 частота выборок / среза)для канала АЦП
float buf1[1];
float process1(register float val) {
  register float tmp, fir, iir;
  tmp = buf1[0];
  iir = val * 0.0304687470912538f;
  iir -= -0.9390625058174924f * tmp;
  fir = tmp;
  fir += iir;
  buf1[0] = iir;
  val = fir;
  return val;
}

void EEPROM_float_write(int addr, float val) // запись в ЕЕПРОМ
{
  byte *x = (byte *)&val;
  for (byte i = 0; i < 4; i++) EEPROM.write(i + addr, x[i]);
}

float EEPROM_float_read(int addr) // чтение из ЕЕПРОМ
{
  byte x[4];
  for (byte i = 0; i < 4; i++) x[i] = EEPROM.read(i + addr);
  float *y = (float *)&x;
  return y[0];
}



//======================================================== Прерывание по нулю
void zero_crosss_int() {

  PORTD &=  ~(1 << PORTD5);

  TCNT1 = 0;

  zero++;

}

//======================================================== Обработка таймера по совпадению нуля
ISR (TIMER1_COMPA_vect)
{
  PORTD |=  (1 << PORTD5);
}

//======================================================== Обработка прерывания АЦП для расчета среднеквадратичного напр.
ISR(ADC_vect)
{
  byte An_pin = ADCL;
  byte An = ADCH;
  Uism = ((An << 8) + An_pin) - 512;
  Uism *= Uism;// возводим значение в квадрат
  Usumm += Uism; // складываем квадраты измерений
  cntr++;
}

void first_start(void) {
  EEPROM_float_write(EE_UST_U_ADDR, DEFAULT_UST_U);
  EEPROM.write(EE_INIT_ADDR, 1);
}

void setup(void) {

  pinMode(detect, INPUT_PULLUP);     //детектор нуля
  pinMode(BUTTON_PLUS, INPUT_PULLUP);    //кнопки, входы
  pinMode(BUTTON_MINUS, INPUT_PULLUP);
  pinMode(BUTTON_SET, INPUT_PULLUP);
  pinMode(ACCELERATOR, INPUT_PULLUP);
  pinMode(STOP, INPUT_PULLUP);
  pinMode(triac, OUTPUT);      //симистор
  lcd.init();
  lcd.backlight();
  if ((EEPROM.read(EE_INIT_ADDR) != 1) || (buffer_ust_U > 265) || (buffer_ust_U < 0)) { //Первый запуск или некорректные значения
    first_start();
  }
  lcd.clear();
  lcd.setCursor(3, 0);
  lcd.print("U REGULATOR");
  lcd.setCursor(3, 1);
  lcd.print("V 0.1");
  _delay_ms(2000);
  lcd.clear();

  //===========================================================Настройка АЦП

  ADMUX = (0 << REFS1) | (1 << REFS0) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0); //

  /*     Биты 7:6 – REFS1:REFS0. Биты выбора опорного напряжения. Если мы будем менять эти биты во время преобразования,
        то изменения вступят в силу только после текущего преобразования. В качестве опорного напряжения может быть выбран AVcc
        (ток источника питания), AREF или внутренний 2.56В источник опорного напряжения.
        Рассмотрим какие же значения можно записать в эти биты:
        REFS1:REFS0
        00    AREF
        01    AVcc, с внешним конденсатором на AREF
        10    Резерв
        11    Внутренний 2.56В  источник, с внешним конденсатором на AREF
        Бит 5 – ADLAR. Определяет как результат преобразования запишется в регистры ADCL и ADCH.
        ADLAR = 0
        Биты 3:0 – MUX3:MUX0 – Биты выбора аналогового канала.
        MUX3:0
        0000      ADC0
        0001      ADC1
        0010      ADC2
        0011      ADC3
        0100      ADC4
        0101      ADC5
        0110      ADC6
        0111      ADC7*/
  //=======================================================================================
  ADCSRA = B11101111; //Включение АЦП
  /*    Бит 7 – ADEN. Разрешение АЦП.
      0 – АЦП выключен
      1 – АЦП включен
      Бит 6 – ADSC. Запуск преобразования (в режиме однократного
      преобразования)
      0 – преобразование завершено
      1 – начать преобразование
      Бит 5 – ADFR. Выбор режима работы АЦП
      0 – режим однократного преобразования
      1 – режим непрерывного преобразования
      Бит 4 – ADIF. Флаг прерывания от АЦП. Бит устанавливается, когда преобразование закончено.
      Бит 3 – ADIE. Разрешение прерывания от АЦП
      0 – прерывание запрещено
      1 – прерывание разрешено
      Прерывание от АЦП генерируется (если разрешено) по завершении преобразования.
      Биты 2:1 – ADPS2:ADPS0. Тактовая частота АЦП
      ADPS2:ADPS0
      000         СК/2
      001         СК/2
      010         СК/4
      011         СК/8
      100         СК/16
      101         СК/32
      110         СК/64
      111         СК/128*/

  //===========================================================
  ACSR = (1 << ACD);
  //------ Timer1 ---------- Таймер задержки времени открытия триака после детектирования нуля (0 триак не откроется)

  TCCR1A = 0x00;  //
  TCCR1B = 0x00;    //
  TCCR1B = (1 << CS12) | (0 << CS11) | (0 << CS10); // Тактирование от CLK.

  // Если нужен предделитель :
  // TCCR1B |= (1<<CS10);           // CLK/1 = 0,0000000625 * 1 = 0,0000000625, 0,01с / 0,0000000625 = 160000 отсчетов 1 полупериод
  // TCCR1B |= (1<<CS11);           // CLK/8 = 0,0000000625 * 8 = 0,0000005, 0,01с / 0,0000005 = 20000 отсчетов 1 полупериод
  // TCCR1B |= (1<<CS10)|(1<<CS11); // CLK/64 = 0,0000000625 * 64 = 0,000004, 0,01с / 0,000004 = 2500 отсчетов 1 полупериод
  // TCCR1B |= (1<<CS12);           // CLK/256  = 0,0000000625 * 256 = 0,000016, 0,01с / 0,000016 = 625 отсчетов 1 полупериод
  // TCCR1B |= (1<<CS10)|(1<<CS12); // CLK/1024
  // Верхняя граница счета. Диапазон от 0 до 255.
  OCR1A = 0;           // Верхняя граница счета. Диапазон от 0 до 65535.

  // Частота прерываний будет = Fclk/(N*(1+OCR1A))
  // где N - коэф. предделителя (1, 8, 64, 256 или 1024)

  TIMSK1 |= (1 << OCIE1A); // Разрешить прерывание по совпадению

  attachInterrupt(1, zero_crosss_int, FALLING);//вызов прерывания при детектировании нуля по спаду

}

void loop(void) {

  //==================================================================================

  if (zero >= 8) {//расчет среднеквадратичного//zero - количество полупериодов дря рассчета среднеквадратичного
    sqrtUsum = Usumm / cntr; //
    real_U = sqrt(sqrtUsum) * 0.725076; // скорректированное 0.7324 из-за питания - оно не 5 В а 4.95, нужно будет подобрать
    real_U = process1(real_U); //фильтруем
    //Serial.println(real_U);
    //Serial.println(cntr);
    //Serial.println("--------------");
    cntr = 0;
    Usumm = 0;// обнуляем суммы напряжений
    zero = 0;
  }

  if (ust_U < 0 || buffer_ust_U < 0) { //Если меньше 0, то 0
    ust_U = 0;
    buffer_ust_U = 0;
  }

  ust_U = (buffer_ust_U > 0) ? buffer_ust_U : 0;

  //=========================================================================Кнопки

  if (digitalRead(STOP) == LOW ) { //считаем сколько времени активен стоп
    if (millis() - time5_press > 1) {
      STOP_press_time++;
      time4_press = millis();
    }
  }

  if ( STOP_press_time > 100) { //если стоп более хх мс, валим напряжение в ноль
    STOP_press_time = 0;
    buffer_ust_U = 0;
    }

    if ( BUTTON_STOP_press_time > 100) { //если комбинация ("-" и "SET" одновременно)более хх мс, тоже валим напряжение в ноль
    BUTTON_STOP_press_time = 0;
    buffer_ust_U = 0;
    }

  if (digitalRead(ACCELERATOR) == LOW ) { //считаем сколько времени активен разгон
    if (millis() - time4_press > 1) {
      ACCELERATOR_press_time++;
      time4_press = millis();
    }
  }

  if ( digitalRead(STOP) == HIGH && ACCELERATOR_press_time > 100) { //если разгон активен более хх мс и нет стопа,
                                                                    //ставим 265 Вольт и взводим флаг разгона
    ACCELERATOR_press_time = 0;
    buffer_ust_U = 265;
    set = 1;
  }

  if (digitalRead(ACCELERATOR) == HIGH && set == 1) {//очищаем счетчик, если разгон отпущен и перед этим был включен, 
                                                     //и на заданное напряжение
    buffer_ust_U = EEPROM_float_read(EE_UST_U_ADDR);
    set = 0;
  }


if (digitalRead(BUTTON_PLUS) == LOW && digitalRead(BUTTON_MINUS) == HIGH && digitalRead(BUTTON_SET) == HIGH) { //считаем сколько времени нажата 1 кнопка
  if (millis() - time1_press > 1) {
    BUTTON_PLUS_press_time++;
    time1_press = millis();
  }
}

if (digitalRead(BUTTON_PLUS) == HIGH && digitalRead(BUTTON_MINUS) == LOW && digitalRead(BUTTON_SET) == HIGH) { //считаем сколько времени нажата 2 кнопка
  if (millis() - time2_press > 1) {
    BUTTON_MINUS_press_time++;
    time2_press = millis();
  }
}

if (digitalRead(BUTTON_PLUS) == HIGH && digitalRead(BUTTON_MINUS) == HIGH && digitalRead(BUTTON_SET) == LOW) { //считаем сколько времени нажата 3 кнопка
  if (millis() - time3_press > 1) {
    BUTTON_SET_press_time++;
    time3_press = millis();
  }
}

if (digitalRead(BUTTON_PLUS) == HIGH && digitalRead(BUTTON_MINUS) == LOW && digitalRead(BUTTON_SET) == LOW) { //считаем сколько времени нажата 2 и 3 кнопка
  if (millis() - time6_press > 1) {
    BUTTON_STOP_press_time++;
    time6_press = millis();
  }
}

if (digitalRead(BUTTON_PLUS) == HIGH) {//очищаем счетчик если кнопка 1 отпущена
  BUTTON_PLUS_press_time = 0;
}

if (digitalRead(BUTTON_MINUS) == HIGH) {//очищаем счетчик если кнопка 2 отпущена
  BUTTON_MINUS_press_time = 0;
}



if ( BUTTON_PLUS_press_time > 100) { //если более хх мс то кнопка 1 нажата, + установки напряжения
  BUTTON_PLUS_press_time = 0;
  buffer_ust_U = buffer_ust_U + 1;
  if (buffer_ust_U > 0 && buffer_ust_U < 10) {
  buffer_ust_U = 10;
  }

  }

if ( BUTTON_MINUS_press_time > 100) { //если более хх мс то кнопка 2 нажата, - установки напряжения
  BUTTON_MINUS_press_time = 0;
  buffer_ust_U = buffer_ust_U - 1;
  if (buffer_ust_U < 10 ) {
    buffer_ust_U = 0;
  }
  }

if ( digitalRead(BUTTON_SET) == HIGH && BUTTON_SET_press_time > 100 && BUTTON_SET_press_time < 1000) { //если кнопку 3 отпустили между 100 и 1000 мс,
                                                                                                       // - установка напряжения из памяти
  BUTTON_SET_press_time = 0;
  buffer_ust_U = EEPROM_float_read(EE_UST_U_ADDR);
  lcd.clear();
  lcd.setCursor(3, 0);
  lcd.print("SETTING");
  _delay_ms(1000);
  lcd.clear();
}
if (digitalRead(BUTTON_SET) == HIGH) {//очищаем счетчик если кнопка 3 отпущена
  BUTTON_SET_press_time = 0;
}

if ( BUTTON_SET_press_time > 1000) { //если кнопку 3 держали дольше1 с,
                                     // - сохранение текущего напряжения в память
  BUTTON_SET_press_time = 0;
  EEPROM_float_write(EE_UST_U_ADDR, buffer_ust_U);
  lcd.clear();
  lcd.setCursor(3, 0);
  lcd.print("SAVING");
  _delay_ms(1000);
  lcd.clear();
}

if (millis() - countMillis >= 500) { //раз в пол секунды регулируем, индицируем
  countMillis = millis();

  step_ = abs (real_U - ust_U);

  if (step_ <= 1) {
    step_ = 1;
  }
  if (step_ >= 40) {
    step_ = 40;
  }


  if (real_U >= (ust_U - 0.3) && real_U <= (ust_U + 0.3))
  {
    angle_U = angle_U;
  }
  if (real_U > (ust_U + 0.3))
  {
    angle_U = angle_U + step_;
  }

  if (real_U < (ust_U - 0.3))
  {
    angle_U = angle_U - step_;
  }

  if (angle_U >= 600) { // ограничил (625 максимум) чтобы не проскакивал целые полупериоды
    OCR1A = 600;
    angle_U = 600;
  }

  if (angle_U <= 5) { // ограничил (0 минимум) чтоб не пропадала напруга на максимуме
    OCR1A = 5;
    angle_U = 5;
  }
  if (ust_U == 0) {
    angle_U = 600;
  }

  if (ust_U != 0) {
    OCR1A = angle_U;
  }

  lcd.setCursor(0, 0);
  lcd.print("OUT U=");
  lcd.print(real_U);
  if (real_U < 100)
  {
    lcd.setCursor(11, 0);
    lcd.print("     ");
  }
  if (real_U < 10)
  {
    lcd.setCursor(10, 0);
    lcd.print("      ");
  }
  lcd.setCursor(12, 0);
  lcd.print("    ");
  lcd.setCursor(0, 1);
  lcd.print("SET U=");
  if (set==1) {
    lcd.print("ACCELERATE");
  }
  else {
  lcd.print(ust_U , 0);
  if (ust_U < 100)
  {
    lcd.setCursor(8, 1);
    lcd.print("        ");
  }
  if (ust_U < 10)
  {
    lcd.setCursor(7, 1);
    lcd.print("         ");
  }
  lcd.setCursor(9, 1);
  lcd.print("       ");
}
}
}






